/*
 * This file contains the class buffered_socket, which provides buffering
 * support for stream sockets.  Presently, only reading actually is buffered;
 * sending calls the underlying socket send call.  The main benefit of read
 * buffering is the ability to implement an efficient readln function, which
 * is useful for old-style line-oriented protocols.  It also can be useful when
 * the application can be most conveniently written using a large number of 
 * small reads.  The buffered_socket will issue large reads and send the
 * data to the client in small peices.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#include "cleansocks.h"
#include "cleansri.h"
#include "cleansrf2sr.h"

#include <string>

#ifndef _cleanbuf_h_
#define _cleanbuf_h_

namespace cleansocks {

class buffered_socket: public srfln_obj {
public:
	static const int DEF_BUF_SIZE = 1024;
	explicit buffered_socket(srf_obj &readfrom, int size = DEF_BUF_SIZE):
		m_socket(&readfrom), m_buf(new char[size]), m_allosize(size),
		m_valid(nullptr), m_validsize(0), m_err_pend(1), m_pending() { }
	explicit buffered_socket():
		m_socket(nullptr), m_buf(nullptr), m_allosize(0),
		m_valid(nullptr), m_validsize(0), m_err_pend(1), m_pending() { }
	void operator()(srf_obj &readfrom, int size = DEF_BUF_SIZE) {
		if(m_socket != nullptr)
			throw cleansocks_reinit_error
				("buffered_socket reinitialized.");
		m_socket = &readfrom;
		m_buf = new char[size];
		m_allosize = size;
	}

	// It is unfortunate that this constructor needs to exist, but
	// I must be able to buffer underlying sockets which don't
	// take flags, in particular cleantlsc.
	explicit buffered_socket(sr_obj &readfrom, int size = DEF_BUF_SIZE):
		m_srf2sr(readfrom),m_socket(&m_srf2sr),
		m_buf(new char[size]), m_allosize(size),
		m_valid(NULL), m_validsize(0), m_err_pend(1), m_pending() { }
	void operator()(sr_obj &readfrom, int size = DEF_BUF_SIZE) {
		if(m_socket != nullptr)
			throw cleansocks_reinit_error
				("buffered_socket reinitialized.");
		m_srf2sr(readfrom);
		m_socket = &m_srf2sr;
		m_buf = new char[size];
		m_allosize = size;
	}

	~buffered_socket() { delete [] m_buf; }
	sr_obj &sock() { return *m_socket; }
private:
	// This allows use with an sr layer below.  Maybe I'll come up with
	// a actually good solution someday.
	srf2sr m_srf2sr;

protected:
	// Underlying socket.
	srf_obj *m_socket;

	// Buffer and allocated size.
	char *m_buf;
	int m_allosize;

	// Area of valid data.
	char *m_valid;
	int m_validsize;

	// Pending system error or EOF.  Positive m_err_pend value is no error,
	// 0 is EOF, -1 is the error given by m_pending.
	int m_err_pend;  // Return value.  0 or -1 indicates "error"
	socket_sys_error m_pending;

	// Send and close are just forwards.
	virtual int send(const void *buf, int size, int flags) override {
		return cleansocks::send(*m_socket,buf,size,clean_flags(flags));
	}
	virtual void close() { cleansocks::close(*m_socket); }

	// Sending isn't modified.
	virtual void shutdownsend() override {
		cleansocks::shutdownsend(*m_socket);
	}

	// Receive needs to discard pending data, then
	// shut the the underlying socket.
	virtual void shutdownrecv() override {
		m_valid = NULL;
		m_validsize = 0;
		cleansocks::shutdownrecv(*m_socket);
	}

	// Receiving uses the buffer.
	virtual int recv(void *buf, int size, int flags) override {
		return recvln(buf, size, -1, flags);
	}

	virtual int recvln(void *buf, int size, int term, int flags);

	// Helper function to copy from the saved buffer.  Return number
	// moved.
	struct unbufret_t {
		int num; bool term;
		unbufret_t(int _num, bool _term): num(_num), term(_term) { }
	};
	unbufret_t unbuffer(char *ubuf, int ubufsize, int term,
			    bool peek = false);

	// Don't even think about it.  Unimplemented.  
	buffered_socket(const buffered_socket&) = delete;
	buffered_socket& operator=(const buffered_socket&) = delete;
};

}
#endif
