/*
 * This file contains a template class to represent a simple integer flag
 * set, but using an integer template type parameter to make the type
 * different from integer, and to create different flag types that differ
 * from each other.  This is all to prevent various far too interesting
 * problems passing an integer flag parameter and having it interpreted
 * as something else.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#ifndef _cleansendopt_h_
#define _cleansendopt_h_

#include <cstring>
#include <functional>
#include <map>

#include "cleansocks.h"

namespace cleansocks {
	
// The send takes the options wrapped as one of these.
template <int Cl>
class clean_flags_tmpl {
public:
	clean_flags_tmpl(): m_flags(0) { }
	explicit clean_flags_tmpl<Cl>(int f): m_flags(f) { }
	explicit operator int() const { return m_flags; }
	operator bool() const { return m_flags != 0; }
	int intval() const { return m_flags; }

	bool operator==(clean_flags_tmpl<Cl> flags) const {
		return m_flags == flags.m_flags;
	}
	bool operator!=(clean_flags_tmpl<Cl> flags) const {
		return m_flags != flags.m_flags;
	}
	clean_flags_tmpl<Cl> & operator=(clean_flags_tmpl<Cl> f) {
		m_flags = f.m_flags;
		return *this;
	}
	clean_flags_tmpl<Cl> & operator=(int f) {
		m_flags = f;
		return *this;
	}
	clean_flags_tmpl<Cl> & operator&=(clean_flags_tmpl<Cl> f) {
		m_flags &= f.m_flags;
		return *this;
	}
	clean_flags_tmpl<Cl> & operator&=(int f) {
		m_flags &= f;
		return *this;
	}
	clean_flags_tmpl<Cl> & operator|=(clean_flags_tmpl<Cl> f) {
		m_flags |= f.m_flags;
		return *this;
	}
	clean_flags_tmpl<Cl> & operator|=(int f) {
		m_flags |= f;
		return *this;
	}
private:
	int m_flags;
};

template <int Cl>
inline clean_flags_tmpl<Cl> operator|(clean_flags_tmpl<Cl> f1,
				      clean_flags_tmpl<Cl> f2) {
	return clean_flags_tmpl<Cl>(static_cast<int>(f1) | static_cast<int>(f2));
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator|(clean_flags_tmpl<Cl> f1, int f2) {
	return clean_flags_tmpl<Cl>(static_cast<int>(f1) | f2);
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator|(int f1, clean_flags_tmpl<Cl> f2) {
	return clean_flags_tmpl<Cl>(f1 | static_cast<int>(f2));
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator&(clean_flags_tmpl<Cl> f1,
				      clean_flags_tmpl<Cl> f2) {
	return clean_flags_tmpl<Cl>(static_cast<int>(f1) & static_cast<int>(f2));
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator&(clean_flags_tmpl<Cl> f1, int f2) {
	return clean_flags_tmpl<Cl>(static_cast<int>(f1) & f2);
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator&(int f1, clean_flags_tmpl<Cl> f2) {
	return clean_flags_tmpl<Cl>(f1 & static_cast<int>(f2));
}
template <int Cl>
inline int operator&=(int f1, clean_flags_tmpl<Cl> f2) {
	return f1 & f2.intval();
}
template <int Cl>
inline int operator|=(int f1, clean_flags_tmpl<Cl> f2) {
	return f1 | f2.intval();
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator^(clean_flags_tmpl<Cl> f1,
				      clean_flags_tmpl<Cl> f2) {
	return clean_flags_tmpl<Cl>(static_cast<int>(f1) ^ static_cast<int>(f2));
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator^(clean_flags_tmpl<Cl> f1, int f2) {
	return clean_flags_tmpl<Cl>(static_cast<int>(f1) ^ f2);
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator^(int f1, clean_flags_tmpl<Cl> f2) {
	return clean_flags_tmpl<Cl>(f1 ^ static_cast<int>(f2));
}
template <int Cl>
inline clean_flags_tmpl<Cl> operator~(clean_flags_tmpl<Cl> f) {
	return clean_flags_tmpl<Cl>(~f.intval());
}

}

#endif

/* Need a compile-time registry for Cl numbers.  Don't know how to do that. */
