/*
 * This contains the Clean Socks declarations for IP 4 networking.  Here are
 * classes for IP 4 sockets, host numbers, port numbers, and endpoints (a
 * pairing of host and port number).  Also methods to look the numbers up from
 * host and port names.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */
#ifndef _cleanip_h_
#define _cleanip_h_

#include <stdint.h>

#include "cleansocks.h"
#include "cleansri.h"

#ifdef WIN32
#include <winsock2.h>
#include <windows.h>
#include <ws2tcpip.h>
typedef uint32_t in_addr_t;
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <iostream>
#include <string>

namespace cleansocks {

// Sockets for us.  Note: since these add no data or methods, but only
// provide constructor parameters, these objects can be assigned to
// to plain socket objects without loss of information or function.
class TCPsocket: public socket, public srf_obj {
public:
	TCPsocket(): socket() { }
	TCPsocket(const socket &s): socket(s, sm_descr) { }
	TCPsocket & operator=(const TCPsocket &s) { 
		fast_assign(s); return *this;
	}
protected:
	virtual int send(const void *buffer, int size, int flags) override;
	virtual int recv(void *buffer, int size, int flags) override;
	virtual void close() override { socket::close(); }
#ifdef WIN32
	virtual void shutdownsend() override { shutdown(SD_SEND); }
	virtual void shutdownrecv() override  { shutdown(SD_RECEIVE); }
#else	
	virtual void shutdownsend() override { shutdown(SHUT_WR); }
	virtual void shutdownrecv() override  { shutdown(SHUT_RD); }
#endif
	virtual sock_descr_t get_sock_descr() const { return sm_descr; }
private:
	// Close one direction.
	void shutdown(int how) {
		if(::shutdown(sock(), how) < 0) {
			int c = ERRCODE;
			throw socket_sys_error("shutdown",c);
		}
	}

	// Socket creation parameters for this class.
	static socket::sock_descr_t sm_descr; // (AF_INET, SOCK_STREAM, "tcp");
};
class UDPsocket: public socket, public strf_obj {
public:
	UDPsocket(): socket() { }
	UDPsocket(const socket &s): socket(s, sm_descr) { }
	UDPsocket & operator=(const UDPsocket &s) { 
		fast_assign(s); return *this;
	}
protected:
	virtual int sendto(const void *buffer, int size, int flags,
			   const endpoint &dest) override;
	virtual int recvfrom(void *buffer, int size, int flags,
			     endpoint *from) override;
	virtual void close() override { socket::close(); }
	virtual sock_descr_t get_sock_descr() const { return sm_descr; }
private:
	// Socket creation parameters for this class.
	static socket::sock_descr_t sm_descr; // (AF_INET, SOCK_DGRAM, "udp");
};

// Error produced by the netdb group.
class socket_db_error: public socket_error {
public:
	// Construct the object using an error code returned by getaddrinfo()
	// that can be interpreted by gai_strerror.
	explicit socket_db_error(std::string id, int code):
		socket_error("[" + id + "] " + gai_strerror(code)),
		m_code(code) { }

	// This will be one of the error codes returned by getaddrinfo and
	// friends.
	int code() const { return m_code; }

	// Check the return code.  If zero, do nothing, otherwise throw a
	// socket_db_error, or socket_sys_error as required.
	static void check_rtn(std::string id, int ret) {
		if(ret == 0) return;
#ifndef WIN32
		if(ret == EAI_SYSTEM) throw socket_sys_error(id, errno);
#endif
		throw socket_db_error(id, ret);
	}
private:
	int m_code;
};

// Object to hold an IP address.
class IPaddress {
public:
	IPaddress(): m_addr(0) { }
	IPaddress(const char *addr) { set(addr); }
	IPaddress(uint8_t w, uint8_t x, uint8_t y, uint8_t z) {
		set(w,x,y,z);
	}
	IPaddress(in_addr_t addr) { set(addr); }

	// Return an IP address with the special wild-card value.
	static IPaddress any() { 
		return IPaddress(static_cast<in_addr_t>(INADDR_ANY)); 
	}

	// Set based on a dotted-decimal string IP address.
	void set(const char *addr) { 
		in_addr_t a = inet_addr(addr); 
		if(a == INADDR_NONE) 
			throw socket_error(std::string("Bad IP address: ") + addr);
		m_addr = a;
	}
	void set(const std::string&s) { set(s.c_str()); }

	// Set from values for each byte.
	void set(uint8_t w, uint8_t x, uint8_t y, uint8_t z) {
		set(256*(256*(256*w+x)+y)+z);
	}

	// Set from a 32-bit integer value (host order).  No bounds check
	// because the arg size is 32 bits.
	void set(in_addr_t addr) { 
		m_addr = htonl(addr); 
	}

	// Set from the value in network order.
	void net(in_addr_t addr) { m_addr = addr; }

	// Get the value.
	in_addr_t get() const { return ntohl(m_addr); }
	in_addr_t net() const { return m_addr; }

	// Get the address as a string.
	std::string str() const;
private:
	friend IPaddress lookup_host(const char *hostname);
	friend class IPendpoint;
	in_addr_t m_addr;
};

// Method to look up and return an IP address.
IPaddress lookup_host(const char *hostname);
inline IPaddress lookup_host(std::string s)
{
	return lookup_host(s.c_str());
}

// Print the IP address.
inline std::ostream & operator<<(std::ostream &strm, const IPaddress &addr) {
	strm << addr.str();
	return strm;
}

// Object to represent an IP port number.
class IPport {
public:
	IPport(): m_port(0) { }
	IPport(uint16_t p) { set(p); }
	IPport(unsigned char high, unsigned char low) { set(high, low); }

	void set(uint16_t p) { m_port = htons(p); }
	void set(unsigned char high, unsigned char low) {
		m_port = htons(256*high + low);
	}
	uint16_t get() const { return ntohs(m_port); }
	uint16_t net() const { return m_port; }
	void net(uint16_t p) { m_port = p; }

	std::string str() const { return std::to_string(ntohs(m_port)); }
private:
	friend IPport lookup_service(const char *);
	friend std::ostream & operator<<(std::ostream &, const IPport &);
	friend class IPendpoint;
	uint16_t m_port;
};

// Lookup a service and return a port.
IPport lookup_service(const char *);
inline IPport lookup_service(std::string s)
{
	return lookup_service(s.c_str());
}

inline std::ostream & operator<<(std::ostream &strm, const IPport &p)
{
	strm << ntohs(p.m_port);
	return strm;
}

// This is the endpoint of a network communication: an IP address and
// an ip port.
class IPendpoint: public endpoint {
public:
	IPendpoint();
	IPendpoint(const IPaddress &addr, const IPport &port);
	IPaddress addr() const {
		IPaddress ret;
		ret.m_addr = m_sa.sin_addr.s_addr;
		return ret;
	}
	IPport port() const {
		IPport ret;
		ret.m_port = m_sa.sin_port;
		return ret;
	}

	std::string str() const { return addr().str() + ":" + port().str(); }
private:
	struct sockaddr_in m_sa;

	friend std::ostream & operator<<(std::ostream &strm,
					 const IPendpoint &p);
};

inline std::ostream & operator<<(std::ostream &strm,
				 const IPendpoint &p)
{
	strm << p.addr() << ":" << p.port();
	return strm;
}

// Sometimes macros, sometimes not.  Causes too much trouble.  Not really
// needed to do
#if 0
// Promote (demote?) these.
#ifdef BSD
#undef htonl
#undef htons
#undef ntohl
#undef ntohs
#endif
inline uint32_t htonl(uint32_t hostlong) { return ::htonl(hostlong); }
inline uint16_t htons(uint16_t hostshort) { return ::htons(hostshort); }
inline uint32_t ntohl(uint32_t netlong) { return ::ntohl(netlong); }
inline uint16_t ntohs(uint16_t netshort) { return ::ntohs(netshort); }


inline uint32_t htonl(const IPaddress &a) { return a.net(); }
inline uint16_t htons(const IPport &p) { return p.net(); }

#endif
}

#endif
