/*
 * This file contains the main classes for the Cleansocks library.  These
 * wrap the socket interface proper, which is abstract and ignorant of the 
 * type of networking.  It contains exception classes for socket errors,
 * a socket class, and endpoint class that represents the sockaddr structure.
 * Since you can't do much with a generic socket, there's not much in the way
 * of operations here, though there are a few implementation-oriented 
 * utilities.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#ifndef _cleansocks_h_
#define _cleansocks_h_

#include <string.h>

#include <iostream>

#include <string>
#include <stdexcept>

#if defined(__NetBSD__) || defined(__APPLE__)
#define BSD
#endif

#ifdef WIN32
#define WINVER 0x0502
#include <winsock2.h>
#include <windows.h>
#define ERRCODE WSAGetLastError()
#else
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#define ERRCODE errno
typedef int SOCKET; // (Stupid windows)
static const int INVALID_SOCKET = -1;
#endif

namespace cleansocks {
	
#ifdef WIN32
typedef int socklen_t;
//#define EWOULDBLOCK WSAEWOULDBLOCK
#define MSG_NOSIGNAL 0
inline char *V2C(void *p) { return static_cast<char *>(p); }
inline const char *V2C(const void *p) { return static_cast<const char *>(p); }
extern int sockinit();
#else
#define V2C(x) x
inline int sockinit() { return 1; }
#endif

// See the end of the file for what this mess is all about.
class socket;
class endpoint;
namespace bindhider {
	inline void bind(const ::cleansocks::socket &s,
			 const ::cleansocks::endpoint &e);
}


// All the exceptions from this package are derived from socket_error, and
// all of those are standard runtime_error.
class socket_error: public std::runtime_error {
public:
	explicit socket_error(const std::string& what): runtime_error(what) { }
};

// These are exceptions generated by failures reported through errno
// on real systems or WSAGetLastError on MS.
class socket_sys_error: public socket_error {
public:
	socket_sys_error(): socket_error("No Error"), m_code(0) { }
	explicit socket_sys_error(int code):
		socket_error(set_code(code)) { }
	explicit socket_sys_error(std::string id, int code):
		socket_error("[" + id + "] " + set_code(code)) { }

	int code() const { return m_code; }

private:
#ifdef WIN32
	std::string set_code(int code);
#else
	std::string set_code(int code)
	{
		char buf[300] = "Unknown Error";
		m_code = code;
#ifdef BSD
		if(strerror_r(m_code, buf, sizeof buf) == 0)
			return buf;
		else
			return "Unknown Error";
#else		
		return strerror_r(m_code, buf, sizeof buf);
#endif		
	}
#endif
	int m_code;
};

// This represents the EAGAIN or EWOULDBLOCK error on non-blocking sockets.
// It is separated from the others for convenience in using non-blocking
// sockets.
class socket_would_block: public socket_sys_error {
public:
	socket_would_block(std::string id, int code): 
		socket_sys_error(id, code) { }
};

class socket_interface_error: public socket_error {
public:
        explicit socket_interface_error(const std::string &s):
		socket_error(s) { }
};


class socket;

// Represents all network enpoint ids.  In TCP, an IP address / port pair.
// This wraps the socket interface sockaddr structure.
class endpoint
{
public:
	// Construct a useless null object.  Need a particular address
	// family to make a useful one.
	endpoint(): m_saptr(NULL), m_size(0) { }

	//sa_family_t family() const { return m_sa.sa_family; }

	const struct sockaddr *addr() const { return m_saptr; }
	struct sockaddr *addr() { return m_saptr; }
	socklen_t addrlen() const { return m_size; }
	socklen_t &addrlen() { return m_size; }
protected:
	// The data area.  The size is the size of the true object.
	struct sockaddr *m_saptr;
	socklen_t m_size;
};

// The socket object.  Conceptually an abstract class, though it is not
// abstract in C++ terms.
class socket {
#ifdef BSD
private:	
	// Now this is truly sorry.  Net/Free BSD does not implement
	// SO_PROTOCOL or SO_DOMAIN in getsockopt.  Not Mac either,
	// per online docs.
	mutable int m_protocol;
	mutable int m_domain;
	static const int BAD_DP = -1;
	void check_broken() const {
		if(m_protocol == BAD_DP || m_domain == BAD_DP)
			throw socket_interface_error
				(std::string("Suboptimal platform: "
					     "does not implment "
					     "SO_PROTOCOL or SO_DOMAIN"));
	}
#else
	void check_broken() const { }
#endif 
public:
	// Various constructors.  Broken BSD must keep the protocol
	// number in the object, so we need to initialize it.
#ifdef BSD
	socket(): m_sock(INVALID_SOCKET),
		  m_domain(BAD_DP), m_protocol(BAD_DP) {}
	socket(int d, int t, int p): m_sock(mksock(d,t,p)),
				     m_domain(d), m_protocol(p) { }
	socket(int d, int t, const char *p): m_domain(d),
		m_sock(mksock(d,t,m_protocol=lookup_proto(p))) { }
	socket(int d, int t, std::string p): m_domain(d),
		m_sock(mksock(d,t,m_protocol=lookup_proto(p.c_str()))) { }
#else
	socket(): m_sock(INVALID_SOCKET) {}
	socket(int d, int t, int p): m_sock(mksock(d,t,p)) { }
	socket(int d, int t, const char *p):
		m_sock(mksock(d,t,lookup_proto(p))) { }
	socket(int d, int t, std::string p):
		m_sock(mksock(d,t,lookup_proto(p.c_str()))) { }
#endif

	socket & operator=(const socket &s) {
		assert_same_descr(s);
		m_sock = s.m_sock;
		return *this;
	}
	
	// Set the type on an untyped socket.  This would only be needed for
	// some specialized app which does not use the standard derived types.
	socket &settype(int d, int t, int p) {
		if(m_sock != INVALID_SOCKET)
			throw socket_interface_error("Cannot re-type a socket");
		m_sock = mksock(d,t,p);
#ifdef BSD
		m_domain = d;
		m_protocol = p;
#endif		
		return *this;
	}
	socket &settype(int d, int t, const char *p) {
		return settype(d,t,lookup_proto(p));
	}
	socket &settype(int d, int t, std::string p) {
		return settype(d,t,p.c_str());
	}
protected:
	// This describes the socket creation type parameters.
	struct sock_descr_t {
		int domain;
		int type;
		int protocol;
		sock_descr_t(int d, int t, int p): 
			domain(d), type(t), protocol(p) { }
		sock_descr_t(int d, int t, const char *p): 
			domain(d), type(t), protocol(lookup_proto(p)) { }
	};

#ifndef WIN32
	// This just simplifies the getsockopt() call for the socket level
	// properties (integers) that we're interested in.  Returns the value
	// or -1.
	int getopt(int opt) const;
#endif //WIN32

	// This returns the description for this socket.  
	virtual sock_descr_t get_sock_descr() const;

	// Throw unless the socket descriptions agree.  
	void assert_descr(const sock_descr_t &descr) const;
	void assert_same_descr(const socket &s) {
		assert_descr(s.get_sock_descr());
	}

	// Copy another socket, but go boom if it's not the right type.
	// Derived classes use this to make safe downcast constructors.
	socket(const socket &s, const sock_descr_t &descr): m_sock(s.m_sock) 
	{ 
		s.assert_descr(descr);
#ifdef BSD
		// Don't think this matters if we pass the assert, but
		// there's nothing there that actually assures that
		// type of *this matches descr, only s.
		m_domain = s.m_domain;
		m_protocol = s.m_protocol;
#endif		
	}

	// Derived classes may call this to avoid the dynamic type check
	// on the assignment operator here.
	void fast_assign(const socket &s) { m_sock = s.m_sock; }

	// Lazy socket allocation.  If called by this class, and the socket
	// is not already typed, the get_sock_descr() will throw.
	// Derived classes which know their descriptions can use this to
	// allocate OS sockets when needed.
	SOCKET sock() const {
		if(m_sock == INVALID_SOCKET) {
			sock_descr_t dsc = get_sock_descr();
			m_sock = mksock(dsc.domain, dsc.type, dsc.protocol);
#ifdef BSD
			m_domain = dsc.domain;
			m_protocol = dsc.protocol;
#endif		
		}
		return m_sock;
	}

	// Basic operations on generic sockets.
	void bind(const endpoint &a) const
	{
		if(::bind(sock(), a.addr(), a.addrlen()) < 0) {
			int c = ERRCODE;
			throw socket_sys_error("bind",c);
		}
	}
	friend void bindhider::bind(const socket &s, const endpoint &a);

	void listen(int backlog) const
	{
		if(::listen(sock(), backlog) < 0) {
			int c = ERRCODE;
			throw socket_sys_error("listen",c);
		}
	}
	friend void listen(const socket &s, int backlog);

	socket accept(endpoint *a) const
	{
		int r = ::accept(sock(), 
				 a ? a->addr() : NULL, 
				 a ? &a->addrlen() : NULL);
		if(r < 0) {
			int c = ERRCODE;
			throw socket_sys_error("accept",c);
		}
#ifdef BSD
		// This assumes the returned socket retains these values.
		// The man page says the allocated socket has "the same
		// properties," which could reasonably mean that.
		return socket(r, "BSD", m_domain, m_protocol);
#else
		return socket(r);
#endif		
	}
	friend socket accept(const socket &s);
	friend socket accept(const socket &s, endpoint &a);

	void connect(const endpoint &a) const
	{
		if(::connect(sock(), a.addr(), a.addrlen()) < 0) {
			int e = ERRCODE;
			throw socket_sys_error("connect",e);
		}
	}
	friend void connect(const socket &s, const endpoint &a);

	void getsockname(endpoint &a) const
	{
		int r = ::getsockname(sock(), a.addr(), &a.addrlen());
		if(r < 0) {
			int c = ERRCODE;
			throw socket_sys_error("getsockname",c);
		}
	}
	friend void getsockname(const socket &s, endpoint &a);

	void close()
	{
		if(m_sock != INVALID_SOCKET)
#ifdef WIN32
			::closesocket(m_sock);
#else
            		::close(m_sock);
#endif
		m_sock = INVALID_SOCKET;
	}
	//friend void close(socket &s);
private:
	// Actual system socket.
	mutable SOCKET m_sock;

#ifdef BSD
	// Constructing one directly is private.  Type info, if
	// provided, should agree with s.  (The string parameter is
	// just to avoid an ambiguity.)
	explicit socket(SOCKET s, std::string st,
			int d, int p):
		m_sock(s), m_protocol(p), m_domain(d) { }
#else				   
	// Constructing one without type info is private.
	explicit socket(SOCKET s): m_sock(s) { }
#endif

	// Look up a protocol string.
	static int lookup_proto(const char *name);

	static SOCKET mksock(int d, int t, int p) {
		SOCKET ret = ::socket(d,t,p);
		if(ret == INVALID_SOCKET)
			throw socket_sys_error("construct socket", ERRCODE);
		return ret;
	}
};

// The *^*^@#&!@*#&_% compiler doesn't like the declaration
//   socket s;
// even under using cleansocks.  Manages to conflict with the system socket
// type.  Must qualify, or use the class keyword.  So here are some aliases.
typedef socket csocket;
typedef socket cleansock;
typedef socket sock;

inline void listen(const socket &s, int backlog = 5)
{
	s.listen(backlog);
}

inline socket accept(const socket &s)
{
	return s.accept(NULL);
}
inline socket accept(const socket &s, endpoint &a)
{
	return s.accept(&a);
}
inline void connect(const socket &s, const endpoint &a)
{
	s.connect(a);
}
inline void getsockname(const socket &s, endpoint &a)
{
	s.getsockname(a);
}
#if 0
// This should exist, but it creates ambiguity when closing sockets which
// are also srxx_objs.  
inline void close(socket &s)
{
	s.close();
}
#endif

/*
 * C++ now contains a std::bind, part of <functional>.  This is a very useful
 * function.  In fact, it is a template.  Unfortunately, when you are using
 * both cleansocks and std namespaces, the unqualified bind() (at least
 * sometimes) matches the std::bind instead of cleansocks::bind.  This seems
 * quite nuts, since the actual bind function should be a better match.  As far
 * as I can tell, the problem has to do with bind being too blasted general,
 * that even converting the parms to const & makes the match less attractive.
 * This is arguably a defect in std::bind, since cleansocks::socket is
 * obviously not a callable object.  Anyway, someone on stackoverflow suggested
 * the following, and it seems to do the job.
 */

// SFINAE version kindly suggested here:
//  https://stackoverflow.com/questions/48073162/im-trying-to-make-a-networking-bind-function-which-matches-before-stdbind
#include <type_traits>
namespace bindhider {
	inline void bind(const ::cleansocks::socket &s,
			 const ::cleansocks::endpoint &e) { s.bind(e); }
}
template<class S0, class S1,
	 typename std::enable_if<std::is_base_of<cleansocks::socket, typename std::decay<S0>::type>::value
			 && std::is_base_of<cleansocks::endpoint, typename std::decay<S1>::type>::value,
		bool>::type = true >
void bind(S0&&s, S1&&e) {
	return bindhider::bind( s, e );
}

}
#endif
