/*
 * Send-recv interface for cleansocks.  This defines the generic interface
 * for using send and recv with the various cleansocks objects.
 *
 * The various usable socket types, and wrappers, implement send() and recv(), 
 * and we want to use inheritance to give different semantics for different
 * types of object.  But we also want to call with external notation, like
 * good ol' plain C.  
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#ifndef _cleansri_h_
#define _cleansri_h_

#include <string>
#include <array>
#include <stdexcept>

#include "cleanflags.h"

namespace cleansocks {

/*
 * Use cleanflags to create a flag set for send and receive.
 */
typedef clean_flags_tmpl<1> clean_flags;

/* Documentation says winsock has MSG_WAITALL, but it's missing when I
   try to build on codeblocks with mingw.  Don't whose fault that is.
   Microsoft's perhaps, for making Windows exist.  Anyway, this file:
     https://github.com/gvanem/wsock-trace/blob/master/src/dump.c
   solves the problem like what I'm doing here.  It might possibly not
   screw things up too badly. */
#if defined(WIN32) && !defined(MSG_WAITALL)
#define MSG_WAITALL 0x8
#endif

//static const clean_flags MSG_DONTWAIT(_cleanflags_value_hack_("MSG_DONTWAIT"));
static const clean_flags CSMSG_DONTROUTE(MSG_DONTROUTE);
static const clean_flags CSMSG_PEEK(MSG_PEEK);
static const clean_flags CSMSG_WAITALL(MSG_WAITALL);

/* For conversion control.  C++ doesn't allow more than one user-defined
   automatic conversions when passing a parameter.  This will allow passing a
   string, but not a char *, which would then require two conversions. */
class stringholder {
private:
	const std::string &m_s;
public:
	stringholder(const std::string &s): m_s(s) { }
	const char *data() const { return m_s.data(); }
	size_t length() const { return m_s.length(); }
};

/* Anything with basic send and recv. */
class sr_obj {
protected:
	virtual int send(const void *buf, int size) = 0;
	virtual int recv(void *buf, int size) = 0;
	virtual void close() = 0;
friend int send(sr_obj &sro, const void *buf, int size);
friend int send(sr_obj &sro, const stringholder &s);
friend int recv(sr_obj &sro, void *buf, int size);
friend void close(sr_obj &sro);
};

/* Anything with send and recv that can have flags.  Also has the 
   shutdown....() methods of TCP fame. */
class srf_obj: public sr_obj {
protected:
	virtual int send(const void *buf, int size, int flags) = 0;
	virtual int send(const void *buf, int size) { return send(buf,size,0); }
	virtual int recv(void *buf, int size, int flags) = 0;
	virtual int recv(void *buf, int size) { return recv(buf,size,0); }
	virtual void shutdownsend() = 0;
	virtual void shutdownrecv() = 0;
friend int send(srf_obj &sro, const void *buf, int size);
friend int send(srf_obj &sro, const void *buf, int size, clean_flags flags);
friend int send(srf_obj &sro, const stringholder &s, clean_flags flags);
friend int recv(srf_obj &sro, void *buf, int size, clean_flags flags);
friend void shutdownsend(srf_obj &sro);
friend void shutdownrecv(srf_obj &sro);
};
	
/* Top-level call wrappers for connected sending. */
inline int send(sr_obj &sro, const void *buf, int size)
{
	return sro.send(buf,size);
}
inline int send(sr_obj &sro, const stringholder &s)
{
	return sro.send(s.data(),s.length());
}
inline int send(srf_obj &sro, const void *buf, int size)
{
	return sro.send(buf,size,0);
}
inline int send(srf_obj &sro, const void *buf, int size, clean_flags flags)
{
	return sro.send(buf,size,flags.intval());
}
inline int send(srf_obj &sro, const stringholder &s, clean_flags flags)
{
	return sro.send(s.data(),s.length(),flags.intval());
}

//Generate an appropriate out_of_range for bad size parm.
inline std::out_of_range oore(std::string op, int size, int limit)
{
	return std::out_of_range
		("cleansocks " + op + " sent size " + std::to_string(size) + 
		 " exceeds array size " + std::to_string(limit));
}
	
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int send(sockish &sro, const std::array<T, N> &buf,
		int size = N, clean_flags flags = clean_flags(0))
{
	if(size > N)
		throw oore("send", size, N);
					
	return cleansocks::send(sro,buf.data(),size,flags);
}
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int send(sockish &sro, const std::array<T, N> &buf,
		clean_flags flags)
{
	return cleansocks::send(sro,buf.data(),N,flags);
}

inline int recv(sr_obj &sro, void *buf, int size)
{
	return sro.recv(buf,size);
}
inline int recv(srf_obj &sro, void *buf, int size, clean_flags flags)
{
	return sro.recv(buf,size,flags.intval());
}
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int recv(sockish &sro, std::array<T, N> &buf,
		int size = N, clean_flags flags = clean_flags(0))
{
	if(size > N)
		throw oore("recv", size, N);
	return cleansocks::recv(sro,buf.data(),size,flags);
}
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int recv(sockish &sro, std::array<T, N> &buf, clean_flags flags)
{
	return cleansocks::recv(sro,buf.data(),N,flags);
}

inline void shutdownsend(srf_obj &sro) {
	sro.shutdownsend();
}
inline void shutdownrecv(srf_obj &sro) {
	sro.shutdownrecv();
}
/* Anything with send, recv and recvln */
class srln_obj: public sr_obj {
protected:
	virtual int recvln(void *buf, int size, int term) = 0;
friend int recvln(srln_obj &sro, void *buf, int size, int term);
};

/* Anything with send, recv and recvln, and also flags on the send and recv,
   and shutdown. */
class srfln_obj: public srf_obj {
protected:
	virtual int recvln(void *buf, int size, int term, int flags) = 0;
	virtual int recvln(void *buf, int size, int term) { 
		return recvln(buf,size,term,0); 
	}
friend int recvln(srfln_obj &sro, void *buf, int size, int term);
friend int recvln(srfln_obj &sro, void *buf, int size, int term,
		  clean_flags flags);
friend int recvln(srfln_obj &sro, void *buf, int size, clean_flags flags);
};

/* External notation wrappers for recvln. */
inline int recvln(srln_obj &sro, void *buf, int size, int term = '\n')
{
	return sro.recvln(buf,size,term);
}
inline int recvln(srfln_obj &sro, void *buf, int size, int term = '\n')
{
	return sro.recvln(buf,size,term,0);
}
inline int recvln(srfln_obj &sro, void *buf, int size, int term,
		  clean_flags flags)
{
	return sro.recvln(buf,size,term,flags.intval());
}
inline int recvln(srfln_obj &sro, void *buf, int size, clean_flags flags)
{
	return sro.recvln(buf,size,'\n',flags.intval());
}
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int recvln(sockish &sro, std::array<T,N> &buf, size_t size,
		  T term = T('\n'), clean_flags flags = clean_flags(0))
{
	if(size > N)
		throw oore("recvln", size, N);
	return cleansocks::recvln(sro,buf.data(),size,term,flags);
}
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int recvln(sockish &sro, std::array<T,N> &buf, T term,
		  clean_flags flags = clean_flags(0)) 
{
	return cleansocks::recvln(sro,buf.data(),N,term,flags);
}
template<typename sockish, typename T, size_t N,
	 typename X = int[3-2*sizeof(T)]>
inline int recvln(sockish &sro, std::array<T,N> &buf,
		  clean_flags flags = clean_flags(0))
{
	return cleansocks::recvln(sro,buf.data(),N,'\n',flags);
}

inline void close(sr_obj &sro)
{
	sro.close();
}

/* So, maybe it's not a stream. */
class endpoint;

class strf_obj {
protected:
	virtual int sendto(const void *buffer, int size, int flags,
			   const endpoint &dest) = 0;
	virtual int recvfrom(void *buffer, int size, int flags,
			     endpoint *from) = 0;
	virtual void close() = 0;
	// Note: from is a pointer, so we can allow NULL.  Don't like that,
	// but it simplifies building the omitted-from-pointer version.

friend int sendto(strf_obj &s, const void *buffer, int size, clean_flags flags,
		  const endpoint &dest);
friend int recvfrom(strf_obj &s, void *buffer, int size, clean_flags flags,
		    endpoint &from);
friend int recvfrom(strf_obj &s, void *buffer, int size, endpoint &from);
friend int recvfrom(strf_obj &s, void *buffer, int size, clean_flags flags);
friend int recvfrom(strf_obj &s, void *buffer, int size);
friend void close(strf_obj &o);
};

/* Top-level call wrappers for message sending. */
inline int sendto(strf_obj &s, const void *buffer, int size, clean_flags flags,
		  const endpoint &dest)
{
	return s.sendto(buffer, size, flags.intval(), dest);
}

inline int sendto(strf_obj &s, const stringholder &b, clean_flags flags,
		  const endpoint &dest)
{
	return sendto(s, b.data(), b.length(), flags, dest);
}
inline int sendto(strf_obj &s, const void *buffer, int size,
		  const endpoint &dest)
{
	return sendto(s, buffer, size, clean_flags(0), dest);
}
inline int sendto(strf_obj &s, const stringholder &b,
		  const endpoint &dest)
{
	return sendto(s, b, clean_flags(0), dest);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int sendto(strf_obj &s, const std::array<T, N> &buffer,
		  int size, clean_flags flags, const endpoint &dest)
{
	if(size > N)
		throw oore("sendto", size, N);
	return cleansocks::sendto(s, buffer.data(), size, flags, dest);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int sendto(strf_obj &s, const std::array<T, N> &buffer,
		  clean_flags flags, const endpoint &dest)
{
	return cleansocks::sendto(s, buffer.data(), N, flags, dest);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int sendto(strf_obj &s, const std::array<T, N> &buffer,
		  int size, const endpoint &dest)
{
	if(size > N)
		throw oore("sendto", size, N);
	return cleansocks::sendto(s, buffer.data(), size, clean_flags(0), dest);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int sendto(strf_obj &s, const std::array<T, N> &buffer,
		  const endpoint &dest)
{
	return cleansocks::sendto(s, buffer.data(), N, clean_flags(0), dest);
}

inline int recvfrom(strf_obj &s, void *buffer, int size, clean_flags flags,
		    endpoint &from)
{
	return s.recvfrom(buffer, size, flags.intval(), &from);
}
inline int recvfrom(strf_obj &s, void *buffer, int size, endpoint &from)
{
	return s.recvfrom(buffer, size, 0, &from);
}
inline int recvfrom(strf_obj &s, void *buffer, int size, clean_flags flags)
{
	return s.recvfrom(buffer, size, flags.intval(), NULL);
}
inline int recvfrom(strf_obj &s, void *buffer, int size)
{
	return s.recvfrom(buffer, size, 0, NULL);
}

template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int recvfrom(strf_obj &s, std::array<T,N> &buf, clean_flags flags,
		    endpoint &from)
{
	return cleansocks::recvfrom(s, buf.data(), N, flags, from);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int recvfrom(strf_obj &s, std::array<T,N> &buf, int size,
		    clean_flags flags, endpoint &from)
{
	if(size > N)
		throw oore("recvfrom", size, N);
	return cleansocks::recvfrom(s, buf.data(), size, flags, from);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int recvfrom(strf_obj &s, std::array<T,N> &buf, int size,
	     endpoint &from)
{
	if(size > N)
		throw oore("recvfrom", size, N);
	return cleansocks::recvfrom(s, buf.data(), size, clean_flags(0), from);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int recvfrom(strf_obj &s, std::array<T,N> &buf, endpoint &from)
{
	return cleansocks::recvfrom(s, buf.data(), N, clean_flags(0), from);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int recvfrom(strf_obj &s, std::array<T,N> &buf,
		    int size = N, clean_flags flags = clean_flags(0))
{
	if(size > N)
		throw oore("recvfrom", size, N);
	return cleansocks::recvfrom(s, buf.data(), size, flags);
}
template<typename T, size_t N, typename X = int[3-2*sizeof(T)]>
inline int recvfrom(strf_obj &s, std::array<T,N> &buf, clean_flags flags)
{
	return cleansocks::recvfrom(s, buf.data(), N, flags);
}

inline void close(strf_obj &o) { o.close(); }

/* This is used by the split-initializaion scheme that tls and buffering,
   and probably other sri classes may use to help with exceptions. */
class cleansocks_reinit_error: public std::runtime_error {
public:
	cleansocks_reinit_error(std::string m): std::runtime_error(m) { }
};
}
#endif
